(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: VolumeTexture.cpp
//
// Desc: Example code showing how to do volume textures in D3D.
//
// Copyright (c) 2000-2001 Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit VolumeTextureUnit;

interface

uses
  Windows, SysUtils,
  Direct3D9, D3DX9, D3DApp, D3DEnumeration, D3DSettings,
  D3DFile, D3DFont, D3DUtil, DXUtil;

//-----------------------------------------------------------------------------
// Defines, constants, and global variables
//-----------------------------------------------------------------------------
type
  TVolumeVertex = packed record
    x, y, z: Single;
    color: DWord;
    tu, tv, tw: Single;
  end;

var
  // this constand initialized in initialization section
  TVolumeVertex_FVF: DWORD = (D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1 {or D3DFVF_TEXCOORDSIZE3(0)});

const
  g_vVertices: array[0..3] of TVolumeVertex = (
    ( x: 1.0; y: 1.0; z: 0.0; color: $ffffffff; tu: 1.0; tv: 1.0; tw: 0.0),
    ( x:-1.0; y: 1.0; z: 0.0; color: $ffffffff; tu: 0.0; tv: 1.0; tw: 0.0),
    ( x: 1.0; y:-1.0; z: 0.0; color: $ffffffff; tu: 1.0; tv: 0.0; tw: 0.0),
    ( x:-1.0; y:-1.0; z: 0.0; color: $ffffffff; tu: 0.0; tv: 0.0; tw: 0.0)
  );

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont: CD3DFont;
    m_pVolumeTexture: IDirect3DVolumeTexture9;
    m_pVB: IDirect3DVertexBuffer9;

  protected
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;

    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function FinalCleanup: HResult; override;
    function Render: HResult; override;
    function FrameMove: HResult; override;

  public
    constructor Create;
  end;

implementation

type
  TVolumeVertexArray = array[0..0] of TVolumeVertex;

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited;

  m_strWindowTitle    := 'VolumeTexture';
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_pFont          := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pVolumeTexture := nil;
  m_pVB            := nil;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  fAngle: Single;
  pVertices: ^TVolumeVertexArray;
  i: Integer;
begin
  fAngle := m_fTime / 2.0;

  // Play with the volume texture coordinate
  m_pVB.Lock(0, 4*SizeOf(TVolumeVertex), Pointer(pVertices), 0);
  for i:= 0 to 3 do
    pVertices[i].tw:= Sin(fAngle)*Sin(fAngle);
  m_pVB.Unlock;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     $00000000, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Draw the quad, with the volume texture
    m_pd3dDevice.SetTexture(0, m_pVolumeTexture);
    m_pd3dDevice.SetFVF(TVolumeVertex_FVF);
    m_pd3dDevice.SetStreamSource( 0, m_pVB, 0, SizeOf(TVolumeVertex));
    m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    // End the scene.
    m_pd3dDevice.EndScene();
  end;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
type
  PDWordArray = ^TDWordArray;
  TDWordArray = array [0..0] of DWord;
var
  LockedBox: TD3DLockedBox;
  w, v, u: Integer;

  pSliceStart: PByte;
  du, dv, dw, fScale: Single;
  r, g, b: DWord;

  pVertices: ^TVolumeVertexArray;
begin
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // Create a volume texture
  Result:= m_pd3dDevice.CreateVolumeTexture(16, 16, 16, 1, 0,
                                            D3DFMT_A8R8G8B8, D3DPOOL_MANAGED,
                                            m_pVolumeTexture, nil);
  if FAILED(Result) then Exit;

  // Fill the volume texture
  {}
    Result:= m_pVolumeTexture.LockBox(0, LockedBox, nil, 0);
    if FAILED(Result) then Exit;

    for w:= 0 to 15 do
    begin
      pSliceStart:= LockedBox.pBits;

      for v:= 0 to 15 do
      begin
        for u:= 0 to 15 do
        begin
          du := (u-7.5)/7.5;
          dv := (v-7.5)/7.5;
          dw := (w-7.5)/7.5;
          fScale:= Sqrt( du*du + dv*dv + dw*dw ) / Sqrt(1.0);

          if (fScale > 1.0) then fScale:= 0.0 else fScale:= 1.0 - fScale;

          r := Trunc((w shl 4)*fScale);
          g := Trunc((v shl 4)*fScale);
          b := Trunc((u shl 4)*fScale);

          PDWordArray(LockedBox.pBits)[u]:= $ff000000 + (r shl 16) + (g shl 8) + (b);
        end;
        LockedBox.pBits:= Pointer(Integer(LockedBox.pBits) + LockedBox.RowPitch);
      end;
      LockedBox.pBits:= Pointer(Integer(pSliceStart) + LockedBox.SlicePitch);
    end;

    m_pVolumeTexture.UnlockBox( 0 );
  {}

  // Create a vertex buffer
  {}
    Result:= m_pd3dDevice.CreateVertexBuffer(4*SizeOf(TVolumeVertex),
                                             D3DUSAGE_WRITEONLY,
                                             TVolumeVertex_FVF,
                                             D3DPOOL_MANAGED, m_pVB, nil);
    if FAILED(Result) then Exit;

    m_pVB.Lock(0, 4*SizeOf(TVolumeVertex), Pointer(pVertices), 0);
    Move(g_vVertices, pVertices^, SizeOf(TVolumeVertex)*4);
    m_pVB.Unlock;
  {}

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  vEye, vAt, vUp: TD3DXVector3;
  matWorld, matView, matProj: TD3DXMatrix;
  fAspect: Single;
begin
  m_pFont.RestoreDeviceObjects();

  // Set the matrices
  vEye:= D3DXVector3(0.0, 0.0,-3.0);
  vAt := D3DXVector3(0.0, 0.0, 0.0);
  vUp := D3DXVector3(0.0, 1.0, 0.0);

  D3DXMatrixIdentity(matWorld);
  D3DXMatrixLookAtLH(matView, vEye, vAt, vUp);
  fAspect:= m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);
  m_pd3dDevice.SetTransform(D3DTS_WORLD,      matWorld);
  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Set state
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_CLIPPING,     iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,     D3DCULL_NONE);
  m_pd3dDevice.SetRenderState(D3DRS_CLIPPING,     iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING,     iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,      iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iFalse);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_SELECTARG1);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pVolumeTexture:= nil;
  m_pVB:= nil;

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  if (0 = (pCaps.TextureCaps and D3DPTEXTURECAPS_VOLUMEMAP)) then Result:= E_FAIL
  else Result:= S_OK;
end;

initialization
  TVolumeVertex_FVF:= (D3DFVF_XYZ or D3DFVF_DIFFUSE or D3DFVF_TEX1 or D3DFVF_TEXCOORDSIZE3(0));
end.

